#ifndef AIFREAD_H_
#define AIFREAD_H_
#include "stdafx.h"
//#include "AvionicsDoc.h"

typedef int vertex;
typedef int edge;
typedef int weight;

enum port_type_t {PUBLISHPORT, SUBSCRIBEPORT, FACET, RECEPTACLE};
enum graph_type_t {PEDG, CEDG, TASKGRAPH};
enum edge_type_t {EVENT, INVOCATION, INTERNAL, REPLICATION};
enum component_type_t {MASTER, PROXY};

#define NUM_RATES 5
enum rate_t {UNASSIGNED, HZ1, HZ5, HZ10, HZ20, HZ40};
//enum consumer_type_t {UNKNOWN, ERM, FULLCHANNEL};
enum color_t {WHITE, GRAY, BLACK};


#define MAX_COMPONENTS 1000
#define MAX_PORTS 10000
#define MAX_COMPONENT_CONN 100000
#define MAX_PORT_CONN 1000000

#define MAX_TASKS 100
#define MAX_TASK_CONN 1000

class CPortNode;
class CComponentNode;
class CTaskNode;
class CProcessorNode;


typedef CTypedPtrList<CPtrList, CComponentNode*> CComponentNodeList;
typedef CTypedPtrList<CPtrList, CPortNode*> CPortNodeList;
typedef CTypedPtrList<CPtrList, CTaskNode*> CTaskNodeList;
typedef CTypedPtrList<CPtrList, CProcessorNode*> CProcessorNodeList;

CDocument *GetDocumentPtr();

class CPortNode {
private:
	CString name;
public:
	int index;
	CString guid;

	bool isTimer;
	bool timerTriggered;
	double wcet; //this should only be valid for subscribe port.

	port_type_t type;
	CComponentNode *parentComp;
	CTaskNode *parentTask;
	rate_t rate;
	bool asynchronous;
	bool correlation;

	CString GetName() { return name;}
	void SetName(const char *name1) { name = name1;}
//	consumer_type_t consumerType;

//	CEventNode *publishedEvent;
//	CEventNodeList subscribedEvents;

	//In the constructor, add myself to my parent's list of ports.
	CPortNode(port_type_t Type, CString &GUID, const char *name1) :
	type(Type), guid(GUID), name(name1), rate(UNASSIGNED), wcet(0,0),
	parentComp(NULL), parentTask(NULL), timerTriggered(false), isTimer(false) {
	}

};

//Instead of having a separate class CTimerNode, I decided to model
//timer as a special type of component, with a single publish port
//that publishes TIMEOUT events.
 
//A component contains a list of ports, which is created and destructed 
//along with the component.
/*class QoSProperties {
public:
	aggregateWCET;
	requiredPeriod;
	expectedPeriod;
	deadline;
	priority;
	importance;
};
*/
class CComponentNode {
public:
	int index;
	CString guid;
	int itemid;
	int groupid;
	
	component_type_t type;
	
	bool isTimer;
	bool timerTriggered;
	rate_t rate;
	bool locked;

	
	CString name;
	CPortNodeList ports;
	CProcessorNode *processor;
	CProcessorNode *processor_allocation;

	//For component allocation, the utilization of the component, sum of all its input ports
	//utils.
	double util;
	
	CComponentNode(const char *Name, component_type_t Type, CString & GUID, int itemid1, int groupid1) :
	name(Name), guid(GUID), timerTriggered(false), rate(UNASSIGNED), processor(NULL), processor_allocation(NULL),
	type(Type), util(0.0), itemid(itemid1), groupid(groupid1), isTimer(false), locked(false) {}
/*
	~CComponentNode() {
		POSITION pos = ports.GetHeadPosition();
		while(pos) {
			CPortNode *port = ports.GetNext(pos);
			delete port;
		}
		ports.RemoveAll();
	}
*/
	CString GetName() { return name;}
	void AddPort(CPortNode *Port) {
		ports.AddTail(Port);
		Port->parentComp = this;

		/*
		Port->SetName(name);
		
		switch (Port->type) {
		case PUBLISHPORT:
			Port->name += ".pub";
			break;
		case SUBSCRIBEPORT:
			Port->name += ".sub";
			break;
		case FACET:
			Port->name += ".facet";
			break;
		case RECEPTACLE:
			Port->name += ".receptacle";
			break;
		}*/
	}
};


class CTaskNode {
public:
	rate_t rate;
	CProcessorNode *parentProcessor;
	CPortNodeList ports;
	//CComponentNodeList components;
	double wcet;
	double wcrt;
	double bcrt;

	int index;
	int priority;

	CTaskNode(rate_t Rate, CProcessorNode *parent) :
	rate(Rate), parentProcessor(parent), wcet(0.0), wcrt(0.0), bcrt(0.0) {}

	void AddPort(CPortNode *Port) {
		ports.AddTail(Port);
		Port->parentTask = this;
	}

	bool ContainsPort(const CPortNode *Port);
	CTaskNode() : wcet(0.0), wcrt(0.0) {}
	CString GetName();
	int GetPeriod();

};

//A processor contains a set of components and a set of tasks.
//A task contains a set of ports
//A component contains a set of ports
//But a task does not contain a component, since a component may be multi-rate.
class CProcessorNode{
public:
	CString name;
	CComponentNodeList components;
	CComponentNodeList components_allocation;

	CTaskNode  tasks[NUM_RATES];
	double cxtSwitchTime;
	CString guid;
	int maxNumComps;

	CProcessorNode(const char *Name, CString &guid1) : name(Name), guid(guid1) {
		tasks[0].rate = HZ1;
		tasks[0].priority = 1;
		tasks[1].rate = HZ5;
		tasks[1].priority = 2;
		tasks[2].rate = HZ10;
		tasks[2].priority = 3;
		tasks[3].rate = HZ20;
		tasks[3].priority = 4;
		tasks[4].rate = HZ40;
		tasks[4].priority = 5;

		cxtSwitchTime = 0.0;
	}
	void CalculateTaskWCET();
	double Interference(const CTaskNode *tNode, double responseTime);
	void RMAnalysis();

	int GetNumTasks();
	CTaskNode *GetMaxPriorityTask();
	bool isSchedulable();
	double GetUtilization();
	double GetUtilBound();

	CTaskNode *GetTaskFromRate(rate_t rate);

	void AddComponent(CComponentNode *cNode) {
		components.AddTail(cNode);
		cNode->processor = this;
	}

	void AddComponentAllocation(CComponentNode *cNode) {
		components_allocation.AddTail(cNode);
		cNode->processor_allocation = this;
	}

	bool ContainsComponent(CComponentNode *cNode1) {
		POSITION pos = components.GetHeadPosition();
		while(pos) {
			CComponentNode *cNode = components.GetNext(pos);
			if(cNode1 == cNode) {
				return true;
			}
		}
		return false;
	}
};

//An integer linked list
///////////////////////////////////////////////////////////////////

class Node
{
public:
    int x; /* The information we really want to store. */
    Node *next;/* The address to the next Node in the list. */
    
    Node() { next = NULL; } /* Start of with no next Node. */
    	~Node() { if (next != NULL) delete next; } 
};

class List
{
   
    Node *head;
   
    Node *tail;
    
    int size;
public:
    List() { size = 0; head = NULL; tail = NULL; }
    List &operator=(List& lst) {
		add(lst);
		return *this;
	}

    ~List() { delete head; }

	void add(List & other);
    void add(int x);
    void del(int x);
	Node *find(int x);
    void print();
	void display(FILE *f);
	void remove_all();

	int get_size() {return size;}
	Node *get_head() { return head;}
	Node *get_next(Node *n) {return n->next;}
};

////////////////////////////////////////////////////////////////////

//Dependency Graph
class wdigraph{
public:	
	graph_type_t type;

	int	N, M;			// max number of vertices and edges
				
	struct wdgedge {
		vertex 	h,t;		// head and tail of the edge
		weight	wt;		// edge weight
		edge	hnext;		// link to next edge incident to head
		edge	tnext;		// link to next edge incident to tail

		CString name;

		edge_type_t type;
		//ACA:
//		bool invalid;
	} *edges;

	void	bldadj();		// build adjacency lists
	void	qsort(int,int);		// quicksort subroutine
	bool	getedge(edge,FILE*);	// get edge from file

	void set_firstout(vertex v, edge e){vertices[v].fout = e;}
	void set_firstin(vertex v, edge e){vertices[v].fin = e;}

public:	
	List DFSPath;
	List BFSSortedNodes;



	void BFSVisit(vertex u);
	void set_edge_type(edge e, edge_type_t type);
	edge_type_t get_edge_type(edge e);
	CString PrintCycleMsg(vertex u);
	void DFSVisitCheckCycles(vertex u);
	void DFSVisitAssignRates(vertex u, rate_t rate);


	struct wdgvertex {
		edge	fout;			// fout[v] is first edge out of v
		edge	fin;			// fin[v] is first edge into v
		color_t color;

		DWORD data;
		
		//ACA
//		weight wt;
//		List merged_vertices;

//		bool invalid;
//		bool is_cluster_head;
	} *vertices;


	wdigraph(graph_type_t Type);
	~wdigraph();
	int	n,m;			// number of vertices and edges

	CString get_edge_name(edge e);
	void set_edge_name(edge e, const char *name);

	
	void assignvertexwts();
	int get_edge_wt(edge e); 
	void set_edge_wt(edge e, int wt);

	CPortNode *get_port_node(vertex v);
	CComponentNode *get_comp_node(vertex v);
	CTaskNode *get_task_node(vertex v);

	void set_rate(vertex v, rate_t rate);

	CString get_node_name(vertex v);

	color_t get_color(vertex v) { return vertices[v].color;}
	void set_color(vertex v, color_t color) {vertices[v].color = color;}
	void reset_color();

	edge	firstout(vertex v) { return vertices[v].fout; }	// return first edge out of v
	edge	firstin(vertex v){ return vertices[v].fin; }	// return first edge into v
	edge	nextout(edge e){ return edges[e].tnext; }		// return next edge out of tail(e)
	edge	nextin(edge e){ return edges[e].hnext; }		// return next edge into head(e)
	vertex	head(edge e){ return edges[e].h; }		// return head of e
	vertex	tail(edge e){ return edges[e].t; }		// return tail of e
	weight	w(edge e){ return edges[e].wt; }		// return weight of e
	
	void    putDotty(const char *fileName);
	void    putDottyAllocation(const char *fileName);

	void dump();
	void reset();
	int join(vertex u, vertex v, weight W = 1);
	edge edge_exists(vertex u, vertex v);


	bool CheckCycles();
	void DFSForwardSlice(vertex u, FILE *f);
	void DFSBackwardSlice(vertex u, FILE *f);

	//ACA: these functions are used to perform a k-way min-cut of the component graph
	//for allocation of components to processors.
//	vertex getmaxvertex();
//	edge getmaxedge(vertex v);
//	void merge(edge e1);
//	void mincut(int numClusters, double maxutil, int maxcomps);
//	wdigraph(const wdigraph &g);

	
	CString getAllocationResultStr();
	
	
//	void bubble_sort(vertex cluster[], int sizeofarray);
	////////////////////////////////////////////////////
	
/* steps in merging and performing the min-cut:
0. make a deep copy of the graph, since the merging algo destroys the graph.
1. mark e as invalid.
2. reposition all edges incoming and outgoing to head(e) to tail(e), add head(e) to
merged_vertices linked list of tail(e). So now head(e) is isolated with no edges incoming
or outgoing, and e is marked as invalid. its like tail(e) absorbed head(e)'s connections, and added
head(e) to its own list.
3. keep going until the required number of clusters is achieved. 
*/

};


class CSystem {
public:
	CComponentNode *GetTimerCompNode(int period);
	CComponentNodeList compNodeList;
	CProcessorNodeList processorNodeList;
	CProcessorNodeList newProcessorNodeList;

	wdigraph thePEDG;
	wdigraph theCEDG;
	wdigraph theTG;

	void Initialize(Udm::SmartDataNetwork *backend);
	void Cleanup();
	void Dump(const char *fileName);
	
	CPortNode *FindPortNodeByGUID(CString &idPort);
	CComponentNode *FindCompNodeByGUID(CString &idComp);
	CProcessorNode *FindProcessorNodeByGUID(CString &id);

	CSystem() : 
		thePEDG(PEDG), theCEDG(CEDG), theTG(TASKGRAPH) {}

	void AssignRatesToPEDG();
	void AssignRatesToCEDG();

	void TimingAnalysis();
};

extern CSystem gTheSystem;
extern CString gEvtsNoSubscriberString;
extern CString gPEDGCycleMsg;
extern CString gUnreachableString;
extern CString gAIFFileNameString;

void ConstructTheSystem(const System& system);
rate_t GetRateFromPeriod(int period);
int GetPeriodFromRate(rate_t rate);

//int GetWeightFromRate(rate_t rate);
CString GetNameFromRate(rate_t rate);
rate_t GetRateFromName(CString name);
CString GetTypeNameFromType(port_type_t type1);
CString GetEdgeNameFromType(edge_type_t type1);

SubscribePort UDMFindSubPortByGUID(Udm::SmartDataNetwork *backend, CString &guid1);
//Facet UDMFindFacetByGUID(Udm::SmartDataNetwork &backend, int guid1);
ComponentInstance UDMFindComponentInstByGUID(Udm::SmartDataNetwork *backend, CString &guid1);
Processor UDMFindProcessorByGUID(Udm::SmartDataNetwork *backend, CString &guid1);

void UpdateAIFWithRateAssignment(Udm::SmartDataNetwork *backend);
void UpdateAIFWithAllocResults(Udm::SmartDataNetwork *backend);
#endif